@@ -1,8 +1,6 @@
 import json
 import os
 from typing import List, Union, Tuple
-
-import pandas as pd
 import requests
 from fhir.resources import FHIRAbstractModel
 from requests import Response
@@ -13,16 +11,15 @@ from fhir.resources.capabilitystatement import CapabilityStatement
 from fhir.resources.reference import Reference
 from requests_oauthlib import OAuth2Session
 import fhir.resources
-from requests_toolbelt import user_agent
-
-from fhir_kindling.fhir_query import FHIRQuery
-from fhir_kindling.fhir_server.auth import generate_auth
-from fhir_kindling.fhir_server.response import ResourceCreateResponse, BundleCreateResponse
-from fhir_kindling.serde import flatten_bundle
+import re
 from oauthlib.oauth2 import BackendApplicationClient
 import pendulum
 
-import re
+from fhir_kindling.fhir_query import FHIRQuery
+from fhir_kindling.fhir_query.query_response import QueryResponse
+from fhir_kindling.fhir_query.query_parameters import FHIRQueryParameters
+from fhir_kindling.fhir_server.auth import generate_auth
+from fhir_kindling.fhir_server.server_responses import ResourceCreateResponse, BundleCreateResponse, ServerSummary
 
 
 class FhirServer:
@@ -30,6 +27,18 @@ class FhirServer:
     def __init__(self, api_address: str, username: str = None, password: str = None, token: str = None,
                  client_id: str = None, client_secret: str = None, oidc_provider_url: str = None,
                  fhir_server_type: str = "hapi"):
+        """
+        Initialize a FHIR server connection
+        Args:
+            api_address: the base endpoint of the fhir server api
+            username: username for basic auth
+            password: password for basic auth
+            token: token for static token auth
+            client_id: client id for oauth2
+            client_secret: client secret for oauth2
+            oidc_provider_url: provider url for oauth2
+            fhir_server_type: type of fhir server (hapi, blaze, etc)
+        """
 
         # server definition values
         self.fhir_server_type = fhir_server_type
@@ -51,7 +60,7 @@ class FhirServer:
         self.session = requests.Session()
         self._setup()
 
-    def query(self, resource: Union[Resource, fhir.resources.FHIRAbstractModel] = None,
+    def query(self, resource: Union[Resource, FHIRAbstractModel, str] = None,
               output_format: str = "json", count: int = 5000) -> FHIRQuery:
         """
         Initialize a FHIR query against the server with the given resource
@@ -80,10 +89,15 @@ class FhirServer:
         Returns:
 
         """
-        valid_query_string, resource = self._validate_query_string_and_parse_resource(query_string)
-        query = FHIRQuery(self.api_address, resource, session=self.session, output_format=output_format)
-        query.set_query_string(valid_query_string)
-        print(query.query_url)
+
+        query_parameters = FHIRQueryParameters.from_query_string(query_string)
+        query = FHIRQuery(
+            self.api_address,
+            resource=query_parameters.resource,
+            query_parameters=query_parameters,
+            session=self.session,
+            output_format=output_format
+        )
         return query
 
     def add(self, resource: Union[Resource, dict]) -> ResourceCreateResponse:
@@ -107,15 +121,22 @@ class FhirServer:
         response = self._upload_resource(resource)
         response.raise_for_status()
 
-        return ResourceCreateResponse(server_response_dict=response.headers, resource=resource)
+        return ResourceCreateResponse(server_response_dict=dict(response.headers), resource=resource)
 
-    def add_all(self, resources: List[Union[Resource, dict]]):
+    def add_all(self, resources: List[Union[Resource, dict]]) -> BundleCreateResponse:
+        """
+        Upload a list of resources to the server, after packaging them into a bundle
+        Args:
+            resources: list of resources to upload to the server, either dictionary or FHIR resource objects
+
+        Returns: Bundle create response from the fhir server
+
+        """
         bundle = self._make_bundle_from_resource_list(resources)
         response = self._upload_bundle(bundle)
         return response
 
     def add_bundle(self, bundle: Union[Bundle, dict, str], validate_entries: bool = True) -> BundleCreateResponse:
-        # todo check this
         # create bundle and validate it
         if isinstance(bundle, dict):
             bundle = Bundle(**bundle)
@@ -139,8 +160,17 @@ class FhirServer:
 
     def delete(self,
                resources: List[Union[Resource, dict]] = None,
-               references: List[Union[str, Reference]] = None):
+               references: List[Union[str, Reference]] = None,
+               query: FHIRQuery = None):
         # todo
+        transaction_bundle = self._make_delete_transaction(resources, references, query)
+
+        response = self.session.post(self.api_address, json=transaction_bundle.dict())
+        response.raise_for_status()
+
+        return response
+
+    def transfer(self, query: FHIRQuery = None, delete: bool = False):
         pass
 
     @classmethod
@@ -166,6 +196,47 @@ class FhirServer:
             else:
                 raise EnvironmentError("Authentication information could not be loaded from environment")
 
+    def _make_delete_transaction(self, resources: List[Union[Resource, dict]] = None,
+                                 references: List[Union[str, Reference]] = None, query: QueryResponse = None) -> Bundle:
+        delete_bundle = Bundle.construct()
+        delete_bundle.type = "transaction"
+        delete_bundle.entry = []
+
+        if resources:
+            if isinstance(resources[0], dict):
+                resources = [fhir.resources.construct_fhir_element(
+                    resource.get("resourceType"),
+                    resource
+                )
+                    for resource in resources
+                ]
+
+            delete_references = [res.relative_path() for res in resources]
+        elif references:
+            if isinstance(references[0], Reference):
+                references = [ref.reference for ref in references]
+            delete_references = references
+        elif query:
+            delete_references = [resource.relative_path() for resource in query.resources]
+        else:
+            raise ValueError("No resources or references provided")
+
+        bundle_entries = [self._make_delete_entry(ref) for ref in delete_references]
+        delete_bundle.entry = bundle_entries
+        return delete_bundle
+
+    @staticmethod
+    def _make_delete_entry(reference: str) -> BundleEntry:
+        entry = BundleEntry.construct()
+        entry.request = BundleEntryRequest(
+            **{
+                "method": "DELETE",
+                "url": reference
+            }
+        )
+
+        return entry
+
     def _make_bundle_from_resource_list(self, resources: List[Union[FHIRAbstractModel, dict]]) -> Bundle:
         upload_bundle = Bundle.construct()
         upload_bundle.type = "transaction"
@@ -206,7 +277,11 @@ class FhirServer:
 
     def _upload_bundle(self, bundle: Bundle) -> BundleCreateResponse:
         r = self.session.post(url=self.api_address, data=bundle.json(return_bytes=True))
-        r.raise_for_status()
+        try:
+            r.raise_for_status()
+        except Exception as e:
+            print(r.text)
+            raise e
         bundle_response = BundleCreateResponse(r, bundle)
         return bundle_response
 
@@ -215,17 +290,8 @@ class FhirServer:
         r = requests.post(url=url, headers=self._headers, auth=self.auth, json=resource.dict())
         return r
 
-    @staticmethod
-    def _format_output(bundle_response: dict, output_format: str) -> Union[dict, pd.DataFrame, Bundle]:
-        if output_format in {"json", "raw"}:
-            return bundle_response
-        elif output_format == "parsed":
-            return Bundle(**bundle_response)
-        elif output_format == "df":
-            return flatten_bundle(bundle_response)
-
     def _get_meta_data(self):
-        url = self.api_address + "/metadata?_format=json"
+        url = self.api_address + "/metadata"
         r = self.session.get(url)
         r.raise_for_status()
         response = r.json()
@@ -239,8 +305,35 @@ class FhirServer:
     def capabilities(self) -> CapabilityStatement:
         if not self._meta_data:
             self._get_meta_data()
-        # todo change back to return capability statement resource
-        return self._meta_data
+        return CapabilityStatement(**self._meta_data)
+
+    @property
+    def rest_resources(self) -> List[str]:
+        return [capa.type for capa in self.capabilities.rest[0].resource]
+
+    def summary(self) -> ServerSummary:
+        summary = self._make_server_summary()
+        return summary
+
+    def _make_server_summary(self) -> ServerSummary:
+        resources = []
+        summary = {
+            "name": self.api_address
+        }
+        for resource in self.rest_resources:
+            url = self.api_address + "/" + resource + "?_summary=count"
+            r = self.session.get(url)
+            r.raise_for_status()
+
+            resource_dict = {
+                "resource": resource,
+                "count": r.json().get("total")
+            }
+            resources.append(resource_dict)
+
+        summary["resources"] = resources
+        summary = ServerSummary(**summary)
+        return summary
 
     @property
     def auth(self) -> Union[requests.auth.AuthBase, None]:
@@ -279,17 +372,6 @@ class FhirServer:
     def _headers(self):
         return {"Content-Type": "application/fhir+json"}
 
-    @staticmethod
-    def _validate_query_string_and_parse_resource(query_string_input: str) -> Tuple[str, FHIRAbstractModel]:
-        if query_string_input[0] != "/":
-            valid_query = "/" + query_string_input
-        else:
-            valid_query = query_string_input
-
-        resource_string = valid_query.split("/")[1].split("?")[0]
-        resource = fhir.resources.get_fhir_model_class(resource_string).construct()
-        return valid_query, resource
-
     @staticmethod
     def validate_api_address(api_address: str) -> str:
         """
@@ -335,6 +417,20 @@ def _api_address_from_env() -> str:
 def _auth_info_from_env() -> Union[str, Tuple[str, str], Tuple[str, str, str]]:
     # First try to load basic auth information
     username = os.getenv("FHIR_USER")
+    # Static token auth
+    token = os.getenv("FHIR_TOKEN")
+    # oauth2/oidc authentication
+    client_id = os.getenv("CLIENT_ID")
+    client_secret = os.getenv("CLIENT_SECRET")
+    oidc_provider_url = os.getenv("OIDC_PROVIDER_URL")
+
+    if username and token:
+        raise EnvironmentError("Conflicting auth information, bother username and token present.")
+    if username and client_id:
+        raise EnvironmentError("Conflicting auth information, bother username and client id present")
+    if token and client_id:
+        raise EnvironmentError("Conflicting auth information, bother static token and client id present")
+
     if username:
         password = os.getenv("FHIR_PW")
         if not password:
@@ -342,22 +438,9 @@ def _auth_info_from_env() -> Union[str, Tuple[str, str], Tuple[str, str, str]]:
         else:
             print(f"Basic auth environment info found -> ({username}:******)")
             return username, password
-    # Static token auth
-    token = os.getenv("FHIR_TOKEN")
-    if username and token:
-        raise EnvironmentError("Conflicting auth information, bother username and token present.")
     if token:
         print("Found static auth token")
         return token
-    # oauth2/oidc authentication
-    client_id = os.getenv("CLIENT_ID")
-    client_secret = os.getenv("CLIENT_SECRET")
-    oidc_provider_url = os.getenv("OIDC_PROVIDER_URL")
-
-    if username and client_id:
-        raise EnvironmentError("Conflicting auth information, bother username and client id present")
-    if token and client_id:
-        raise EnvironmentError("Conflicting auth information, bother static token and client id present")
 
     if client_id and not client_secret:
         raise EnvironmentError("Insufficient auth information, client id specified but no client secret found.")
